unit ADF_file_routines;

interface
uses
  SysUtils, Classes, Dialogs;

const


    // 512*11*2*80 = 901120 bytes on a disk

    // There 1760 (901120/512) blocks on a disk
    // Two blocks are taken up by the BootBlock
    // 1 block is taken up by the RootBlock
    // All the other blocks are used for files etc..

    BSIZE             = 512; // Block/Sector Size (1 sector = 1 block)
    SECTORS_PER_TRACK = 11;
    TRACKS_PER_CYL    = 2;
    NUM_CYLS          = 80;

    ADF_BUFFER_SIZE = BSIZE * SECTORS_PER_TRACK *
                      TRACKS_PER_CYL * NUM_CYLS; // Amount of bytes a whole ADF image takes

type
  //===========================================================================
  // Records to hold various information about the ADF image
  // 1 byte = 8 bits
  // 1 word = 2 bytes
  // 1 longword = 2 words or 4 bytes
  //===========================================================================

  TbootBlock = RECORD
    disktype        :array[0..2] of char;     // 'DOS' or 'PFS'   (3 bytes)
    disktype_flags  :byte;                    // OFS, FFS etc..   (1 byte)
    checksum        :longword;                // Block checksum   (4 bytes)
    rootblock       :longword;                // Ptr to RootBlock (4 bytes)
    bootblock_code  :array[0..1011] of byte;  // BootBlock code   (1012 bytes)
                                              //       Total      (1024 bytes)
  end;

  TFloppyRootBlock = RECORD
    primtype        :longword;    // Value = $00000002 (T_HEADER)
    header_key      :longword;    // Unused in Rootblock (val = $00000000)
    high_seq        :longword;    // Unused (val = $00000000)
    hashtable_size  :longword;    // for floppy disk this value = $00000048
    first_data      :longword;    // unused
    checksum        :longword;    // Block checksum
    hashtable       :array[0..71] of longword;  // 72 longwords
    bm_flag         :longword;    // -1 means VALID
    bm_pages        :array[0..24] of longword;  // bitmap blocks pointers
    bm_ext          :longword;    // First bitmap extension block (Hard disk only)
    r_days          :longword;    // Last root alteration date, days since 01/01/1978
    r_mins          :longword;    // Minutes past midnight
    r_ticks         :longword;    // ticks (1/50sec) past last minute
    name_len        :byte;        // Volume name length
    diskname        :array[0..29] of char;
    unused1         :char;
    unused2         :array[0..1] of longword;

    v_days          :longword;    // Date when disk was last written to
    v_mins          :longword;
    v_ticks         :longword;

    c_days          :longword;    // Date when disk was last formatted
    c_mins          :longword;
    c_ticks         :longword;

    next_hash       :longword;    // Unused
    parent_dir      :longword;    // Unused
    extension       :longword;    // FFS = first directory cache block, 0 otherwise
    sec_type        :longword;    // Block secondary type = ST_ROOT (value 1)
  end;

  TUserDirectoryBlock = RECORD
    primtype        :longword;    // Value = $00000002 (T_HEADER)
    header_key      :longword;
    unused1         :array[0..2] of longword;
    checksum        :longword;
    hashtable       :array[0..71] of longword;  // 72 longwords
    unused2         :longword;
    user_id         :word;        // User ID
    group_id        :word;        // Group ID
    protect_flags   :longword;    // Protection flags (set to 0 by default)
    unused3         :longword;
    comment_len     :char;        // Directory comment length
    comment         :array[0..78] of char;
    unused4         :array[0..11] of char;
    days            :longword;
    mins            :longword;
    ticks           :longword;
    name_len        :char;        // Directory name length
    dirname         :array[0..29] of char;
    unused5         :char;
    unused6         :array[0..1] of longword;
    next_link       :longword;
    unused7         :array[0..4] of longword;
    hash_chain      :longword;
    parent          :longword;
    extension       :longword;
    sec_type        :longword;
  end;


  TFileHeaderBlock = RECORD
    primtype        :longword;    // Value = $00000002 (T_HEADER)
    header_key      :longword;    // Self pointer (to this block)
    high_seq        :longword;    // Number of data block ptr stored here
    data_size       :longword;    // Unused
    first_data      :longword;    // First data block ptr
    checksum        :longword;    // Same algorithm as RootBlock
    data_blocks     :array[0..71] of longword;
    unused1         :longword;
    user_id         :word;        // User ID
    group_id        :word;        // Group ID
    protect_flags   :longword;    // Protection flags (set to 0 by default)
    byte_size       :longword;    // File size in bytes
    comm_len        :char;        // File comment length (max 79)
    comment         :array[0..78] of char;  // Comment   (max 79 chars)
    unused2         :array[0..11] of char;        // Set to 0
    days            :longword;    // Last change date of file
    mins            :longword;
    ticks           :longword;
    name_len        :char;        // Filename length
    filename        :array[0..29] of char;   // Max 30 chars permitted
    unused3         :char;
    unused4         :longword;
    real_entry      :longword;    // FFS : unused (=0)
    next_link       :longword;    // FFS : hardlinks chained list (first=newest)
    unused5         :array[0..4] of longword;    // set to 0
    hash_chain      :longword;    // Next entry ptr with same hash
    parent          :longword;    // Parent directory
    extension       :longword;    // Pointer to 1st file extension block
    sec_type        :longword;    // Secondary type : ST_FILE (= -3)
  end;

  TBitMapBlock = RECORD
    checksum        :longword;
    map             :array[0..126] of longword;
  end;

  TFileDetails = class(TObject)
    name            :String;
    parentdir       :String;
    size            :longword;
    headerblock     :longword;
  end;



type
  TADFlib = class(TObject)

  private

    // Complete disk image stored as bytes in here.
    sectorMap           :array[0..ADF_BUFFER_SIZE-1] of byte;
    bootBlockPtr        :^TBootBlock;       // Pointer to Boot Block
    rootBlockPtr        :^TFloppyRootBlock; // Pointer to Root Block

    // Various internal status variables
    fADF_file           :File;
    fFileName           :String;
    fFileLoaded         :Boolean;



    procedure generateDirectoryList(blockNum: Longword);
    procedure FlipLong(var l           :longword);
    function CalcChecksum(var X; XSize :longword; CheckSumPos:longword; Method:longword):longword;

  public

    ADFdirectoryList    :TStringList;
    ADFdirectoryListCnt :Integer;

    constructor Create;
    destructor Destroy; override;

    property FileLoaded   :Boolean read fFileLoaded;
    property filesInADF   :Integer read ADFDirectoryListCnt;

    procedure loadADF(fname: String);     // Load an ADF file
    procedure unLoadADF;                  // unLoad an ADF file

    function getWindowsFileName :String;  // Returns 'Windows' filename of ADF
    function getDiskLabel       :String;  // Returns Disk Name/Label
    function getFileSystem      :String;  // Returns Filesystem  (DOS,PFS etc..)
    function getDiskType        :String;  // Returns Disk Type   (OFS,FFS etc..)

    function calcBlockChecksum(blockNum :longword) :String;
    function getBlockChecksum(blockNum  :longword) :String;
    function getBlockSecType(blockNum   :longword) :String;
    function getParentDirName(blockNum  :longword) :String;
    function getFileSizeBytes(blockNum  :longword) :longword;
    function isBlockFree(blockNum       :longword) :Boolean;

    procedure extractFile(blockNum      :longword);

    // Functions that retrieve data from the TFileDetails object
    function listFile(idx  :Integer): String;

    // Worker functions
    function getByteAsChr(pos: longint): String;
    function getByteAsHex(pos: longint): String;


  end;




implementation

{==========================================================================
 Constructor
==========================================================================}
constructor TADFlib.Create;
begin
  unLoadADF;    // Set some initial default values

  // Pointers to various structures in the disk sector map
  bootBlockPtr := @sectorMap[0];
  rootBlockPtr := @sectorMap[880*512];   // block 880
end;

{==========================================================================
 Destructor
==========================================================================}
destructor TADFlib.Destroy;
begin
  inherited Destroy;
  unLoadADF;
end;


{###########################################################################
  FUNCTIONS IN THIS SECTION ARE WINDOWS HOUSEKEEPING
  IE: LoadADF, UnLoadADF, getWindowsFilename etc...
###########################################################################}

{==========================================================================
 Function:      LoadADF
 Description:   Loads a complete ADF image into sectorMap array

 ACCEPTS:       Filename
 RETURNS:       Nothing
==========================================================================}
procedure TADFlib.loadADF(fname: String);
var
  fFloppySize     :Longint;

begin
  // If an existing file has already been opened then user must either
  // Save or close it before opening this one.

  if (fFileLoaded = true) then
    // Unload any details about the previous ADF before loading this new one.
    unLoadADF;

  // Load new ADF details
  fFileName     := fname;

  fillchar(sectorMap, ADF_BUFFER_SIZE, $00);// Zero the array
  AssignFile(fADF_file, fname);             // Read the file into memory
  FileMode := fmShareDenyNone;              // Allow others to access file
  reset(fADF_File, 1);                      // Open the file
  fFloppySize := FileSize(fADF_file);       // Get filesize...
  Seek(fADF_file, 0);                       // Seek to beginning of file

  // Dont read past end of buffer size
  if (fFloppySize > ADF_BUFFER_SIZE) then
    fFloppySize := ADF_BUFFER_SIZE;

  if (fFloppySize > 0) then
    BlockRead(fADF_file, sectorMap, fFloppySize); // Read nnn bytes of data

  CloseFile(fADF_file);
  fFileLoaded := true;

  // These have to be here and not in the generateDirectoryList
  // routine as it is self calling recursive routine
  ADFdirectoryList    := TStringList.create;
  ADFdirectoryListCnt := 0;
  generateDirectoryList(880);
  ADFdirectoryList.Sort;

end;

{==========================================================================
 Function:      UnLoadADF
 Description:   UnLoads an ADF image

 ACCEPTS:       Nothing
 RETURNS:       Nothing
==========================================================================}
procedure TADFlib.unLoadADF;
var
  i   :Integer;
begin
  fillchar(sectorMap, ADF_BUFFER_SIZE, $00);    // Zero the array
  fFileLoaded := false;
  fFileName   := '';

  // Clear the directory list
  if ADFDirectoryList <> nil then
    // Destroy all the objects within the directory list first
    for i := 0 to ADFDirectoryList.Count - 1 do
      ADFDirectoryList.Objects[i].Free;

    FreeAndNil(ADFDirectoryList);
  ADFDirectoryListCnt := 0;
end;


{==========================================================================
 Function:      getWindowsFileName
 Description:   Returns Windows Filename of ADF image

 ACCEPTS:       Nothing
 RETURNS:       Filename
==========================================================================}
function TADFlib.getWindowsFileName: String;
begin
  if (fFileLoaded) then
    Result := ExtractFileName(fFileName)
  else
    Result := 'No File Loaded';
end;


{###########################################################################
  FUNCTIONS IN THIS SECTION RETURN GENERIC DISK INFORMATION
  Amiga Disk Name (Label)
  Disk File System (DOS, PFS)
  Disk File Type   (OFS, FFS etc..)
###########################################################################}

{==========================================================================
 FUNCTION:    getDiskLabel
 DESCRIPTION: Returns the Amigados name of the disk

 ACCEPTS:     Nothing
 RETURNS:     Nothing
==========================================================================}
function TADFlib.getDiskLabel: String;
var
  i           :Integer;
  diskLabel   :String;

begin
  if (fFileLoaded) then
  begin
    rootBlockPtr := @sectorMap[880*512];

    for i := 0 to rootBlockPtr.name_len-1 do
    begin
      diskLabel := diskLabel + rootBlockPtr.diskname[i];
    end;
    Result := diskLabel;
  end
  else
  begin
    Result := 'No File Loaded';
  end;
end;

{==========================================================================
 Function:      getFileSystem
 Description:   Returns File System of ADF image (DOS, PFS etc..)

 ACCEPTS:       Nothing
 RETURNS:       File System
==========================================================================}
function TADFlib.getFileSystem: String;
begin
  if (fFileLoaded) then
    Result := bootBlockPtr.disktype
  else
    Result := 'No File Loaded';
end;

{==========================================================================
 FUNCTION:    getDiskType
 DESCRIPTION: Returns type of disk

 ACCEPTS:     Nothing
 RETURNS:     Type of disk
==========================================================================}
function TADFlib.getDiskType: String;
var
   dsktype   :byte;
begin

  if (fFileLoaded) then
  begin
    dskType := bootBlockPtr.disktype_flags;
    case dskType of
      0 : Result := 'OFS';
      1 : Result := 'FFS';
      2 : Result := 'OFS/INTL';
      3 : Result := 'FFS/INTL';
      4 : Result := 'OFS/DIRC&INTL';
      5 : Result := 'FFS/DIRC&INTL';
    end;
  end
  else
    Result := 'No File Loaded';
end;


{###########################################################################
  FUNCTIONS IN THIS SECTION NEED BLOCK NUMBER PASSED IN
###########################################################################}

{==========================================================================
 FUNCTION:    calcBlockChecksum
 DESCRIPTION: Calc the correct Checksum of a Block - calls Feltz's routine

 ACCEPTS:     blockNum
 RETURNS:     Checksum as a string.
==========================================================================}
function TADFlib.calcBlockChecksum(blockNum  :Longword) :String;
var
  fCheckSum   :longword;
begin
  if (fFileLoaded) then
  begin
    if (blockNum = 0) then
      // Calc bootblock checksum
      fChecksum := CalcChecksum(sectorMap[blockNum*512],1024,4,0)
    else
      // Calc other block checksum
      fChecksum := CalcChecksum(sectorMap[blockNum*512],1024,4,1);

    Fliplong(fCheckSum);
    Result := Format('$%0.8x', [fCheckSum]);
  end
  else
      Result := 'No File Loaded';
end;


{==========================================================================
 FUNCTION:    getBlockChecksum
 DESCRIPTION: Get the Checksum of a Block - Checksum can be wrong

 ACCEPTS:     BootBlock, OtherBlock
 RETURNS:     Checksum
==========================================================================}
function TADFlib.getBlockChecksum(blockNum  :longword) :String;
var
  fOtherBlock :^TFloppyRootBlock;

  fCheckSum   :longword;
begin
  if (fFileLoaded) then
  begin
    if (blockNum = 0) then
    begin
      // Get bootblock checksum
      fCheckSum  := bootBlockPtr.checksum;
    end
    else
    begin
      // Get any other block's checksum
      fOtherBlock := @sectorMap[blockNum*512];
      fcheckSum   := fOtherBlock.checksum;
    end;

    Fliplong(fChecksum);
    Result := Format('$%0.8x', [fChecksum]);
  end
  else
      Result := 'No File Loaded';
end;


{==========================================================================
 FUNCTION:    getBlockSecType
 DESCRIPTION: Returns

 ACCEPTS:     Block Number
 RETURNS:     Type of Block (Root, File Header, Directory etc..)
==========================================================================}
function TADFlib.getBlockSecType(blockNum :longword): String;
type
  TBlockSecType = RECORD
    unused          :array[0..126] of longword; // Ignore first 508 bytes
    sec_type        :longword;                  // Block secondary type
  end;

var
  secType          :longword;
  BlockSecTypePtr  :^TBlockSecType;

begin

  if (fFileLoaded) then
  begin
    BlockSecTypePtr := @sectorMap[blockNum * 512];
    secType := BlockSecTypePtr.sec_type;
    Fliplong(secType);

    case secType of
      $00000001 : Result := 'Root';
      $00000002 : Result := 'Directory';
      $00000003 : Result := 'Soft Link';
      $00000004 : Result := 'Hard Link';
      $FFFFFFFD : Result := 'File';       // -3
      $FFFFFFFC : Result := 'Hard Link';  // -4
    else
      if (blockNum = 0) or (blockNum = 1) then
        Result := 'Boot'
      else
        Result := 'Data';
    end;
  end
  else
  begin
    Result := 'No File Loaded';
  end;

end;


{==========================================================================
 FUNCTION:    isBlockFree
 DESCRIPTION: Checks if a block has been used or not

 ACCEPTS:     Block Number
 RETURNS:     TRUE if block is Free / FALSE if block is Used (allocated)
==========================================================================}
function TADFlib.isBlockFree(blockNum: longword): Boolean;
var
  arrpos          :Longword;
  diff            :Longword;
  bit             :Longword;
  shifted         :Longword;

  bitMapBlockPtr  :Longword;
  fRootBlock      :^TFloppyRootBlock;
  fBitMapBlock    :^TBitMapblock;

begin

  // The Root block bm_pages[] array contains pointers to one or more
  // bitmap blocks.
  // Each bitmap block contains details about whether all blocks on the disk
  // are free or allocated (starting with block 2 as blocks 0 and 1 are
  // Bootblock blocks and are ignored)
  // There are 1760 blocks in an adf disk so the bitmap pages array will
  // be 1758 bits long (54 longwords + 30 bits of the 55th longword)
  // If a bit is set then the block is free

  if (fFileLoaded) then
  begin

    // Block 2 is stored at bit 0 so account for the shift
    if (blockNum < 2) or (blockNum > 1760) then
    begin
      showMessage('isBlockFree: blockNum '+IntToStr(blockNum)+' out of range');
      Result := true;
    end
    else
    begin
      // As we are only interested in ADF's then we only need to jump to
      // the bitmap block located in bm_pages[0]
      fRootBlock := @sectorMap[880*512];
      bitMapBlockPtr := fRootBlock.bm_pages[0];
      Fliplong(bitMapBlockPtr);
      fBitMapBlock := @sectorMap[bitMapBlockPtr*512];

      // Read the correct bit for the block Number we want to check
      blockNum := blockNum - 2;
      arrpos := blockNum div 32;          // arrpos = correct longword in list
      diff   := blockNum - (arrpos * 32); // diff = correct bit in longword
      shifted := 1 shl diff;              // Shifted = position of set bit for '&' mask

      bit := fBitMapBlock.map[arrpos];
      Fliplong(bit);

      bit := bit and shifted;

      if bit = 0 then
        Result := False   // block is Used
      else
        Result := True;   // block is Free
    end;
  end
  else
  begin
    Result := true;       // Error so set as Used
  end;
end;



{###########################################################################
  FUNCTIONS IN THIS SECTION WORK ON AMIGA DIRECTORIES WITHIN THE ADF
###########################################################################}


{==========================================================================
 FUNCTION:    getParentDirName
 DESCRIPTION: Returns the parentDirectory of a Directory

 ACCEPTS:     Block Number (must be a User Directory Block)
 RETURNS:     Name of parent Directory
==========================================================================}
function TADFlib.getParentDirName(blockNum: longword): String;
var
  fileBlock     :^TFileHeaderBlock;
  uDirBlock     :^TUserDirectoryBlock;

  secType            :String;
  parentDirBlockNum  :longword;
begin

  secType := getBlockSecType(blockNum);

  if (secType = 'File') then
  begin
    fileBlock := @sectorMap[blockNum*512];
    parentDirBlockNum := fileBlock.parent;
  end
  else
  begin
    uDirBlock := @sectorMap[blockNum*512];
    parentDirBlockNum := uDirBlock.parent;
  end;

  Fliplong(parentDirBlockNum);

  if (parentDirBlockNum = 0) or (parentDirBlockNum = 880) then
    Result := ''
  else
  begin
    uDirBlock := @sectorMap[parentDirBlockNum*512];
    Result := uDirBlock.dirname;
  end;
end;


{###########################################################################
  FUNCTIONS IN THIS SECTION WORK ON AMIGA FILES WITHIN THE ADF
###########################################################################}

{==========================================================================
 FUNCTION:    getFileSizeBytes
 DESCRIPTION: Returns the Filesize of a given file in bytes

 ACCEPTS:     Block Number (must be a File Header Block)
 RETURNS:     Filesize in bytes (0 if not a File Header block)
==========================================================================}
function TADFlib.getFileSizeBytes(blockNum: longword): longword;
var
  fileBlock     :^TFileHeaderBlock;
  f             :longword;
begin
  if (getBlockSecType(blockNum) = 'File') then
  begin
    fileBlock := @sectorMap[blockNum*512];
    f := fileBlock.byte_size;
    Fliplong(f);
    Result := f;
  end
  else
  begin
    showMessage('getFileSizeBytes(): Block '+IntToStr(blockNum)+' is not a file header block');
    Result := 0;
  end;
end;


{==========================================================================
 FUNCTION:    listFile
 DESCRIPTION: Returns a filename from the ADFdirectory  TStringList

 ACCEPTS:     index value
 RETURNS:     Full path and filename
==========================================================================}
function TADFlib.listFile(idx: Integer): String;
begin
  if (idx >= 0) and (idx < ADFdirectoryListCnt) then
    Result := ADFdirectoryList[idx]
  else
    Result := 'listFile('+IntToStr(idx)+') out of bounds';
end;


{==========================================================================
 FUNCTION:    extractFile
 DESCRIPTION: Saves a chosen file within the adf image to the local harddisk

 ACCEPTS:     Block Number (must be a File Header Block)
 RETURNS:     Nothing.
==========================================================================}
procedure TADFlib.extractFile(blockNum: longword);
var
  fileBlock         :^TFileHeaderBlock;
  num_bytes_to_save :longword;

begin

  if (getBlockSecType(blockNum) = 'File') then
  begin
    fileBlock := @sectorMap[blockNum*512];
    num_bytes_to_save := fileBlock.byte_size;
    Fliplong(num_bytes_to_save);

  end
  else
  begin
    showMessage('extractFile(): Block '+IntToStr(blockNum)+' is not a file header block');
  end;
end;


{###########################################################################
  ANY MISC FUNCTIONS GO IN THIS SECTION
###########################################################################}

{==========================================================================
 FUNCTION:    getByteAsChr
 DESCRIPTION: Returns byte at pos as a chr
 ACCEPTS:     Offset into sectorMap array
 RETURNS:     Byte value as a chr
==========================================================================}
function TADFlib.getByteAsChr(pos: longint): String;
begin
  Result := chr(sectorMap[pos]);
end;

{==========================================================================
 FUNCTION:    getByteAsHex
 DESCRIPTION: Returns byte at pos as two digit Hex value
 ACCEPTS:     Offset into sectorMap array
 RETURNS:     Byte value as a Hex String
==========================================================================}
function TADFlib.getByteAsHex(pos: longint): String;
begin
  Result := Format('%0.2x', [sectorMap[pos]]);
end;



{###########################################################################
  PRIVATE FUNCTIONS IN THIS SECTION
###########################################################################}

{==========================================================================
 FUNCTION:    generateDirectoryList
 DESCRIPTION: Gets the list of files within the ADF and puts them in
              the ADFdirectoryList  TStringList
 ACCEPTS:     longword value
 RETURNS:     flipped longword value
==========================================================================}
procedure TADFlib.generateDirectoryList(blockNum: Longword);
var
  i                 :Integer;
  secType           :String;
  parent_dir        :String;
  dir_or_file_name  :String;
  nextChainPtr      :Longword;
  nextDirPtr        :Longword;
  fileBlock         :^TFileHeaderBlock;
  uDirBlock         :^TUserDirectoryBlock;

  fDetails          :TFileDetails;

  //myTest            :TFileDetails;

begin
  secType := getBlockSecType(blockNum);

  if ((secType = 'Directory') or (secType = 'Root')) then
  begin
    uDirBlock        := @sectorMap[blockNum*512];
    dir_or_file_name := uDirBlock.dirname;
    parent_dir       := getParentDirName(blockNum) + '/';

    // As we have identified a directory we need to recursively delve into
    // it and retrieve its files and sub-dirs.
    for i := 0 to 71 do
    begin
      nextDirPtr := uDirBlock.hashtable[i];
      FlipLong(nextDirPtr);

      if nextDirPtr <> 0 then
      begin
        //showMessage(Format('$%0.8x', [fBlockNum]));
        generateDirectoryList(nextDirPtr);
      end;
    end;

    // Is there any other directory with the same hash value as this one?
    nextChainPtr := uDirBlock.hash_chain;
    Fliplong(nextChainPtr);

    if (nextChainPtr <> 0) then
    begin
      generateDirectoryList(nextChainPtr);
    end;
  end;

  if (secType = 'File') then
  begin
    fileBlock        := @sectorMap[blockNum*512];
    dir_or_file_name := fileBlock.filename;
    parent_dir       := getParentDirName(blockNum) + '/';

    // Is there any other file with the same hash value as this one?
    // If so, then go get that file as well.
    nextChainPtr := fileBlock.hash_chain;
    FlipLong(nextChainPtr);

    if (nextChainPtr <> 0) then
      generateDirectoryList(nextChainPtr);

    fDetails             := TFileDetails.create;
    fDetails.name        := dir_or_file_name;
    fDetails.parentdir   := parent_dir;
    fDetails.size        := fileBlock.byte_size;
    fDetails.headerblock := blockNum;

    ADFdirectoryList.AddObject(parent_dir + dir_or_file_name, fDetails);
    //myTest := ADFdirectoryList.Objects[ADFdirectoryListCnt] as TFileDetails;
    //showMessage(myTest.name+'  '+IntToStr(myTest.headerblock));
    Inc(ADFdirectoryListCnt);
  end;
end;


{==========================================================================
 FUNCTION:    Fliplong
 DESCRIPTION: Flips a longword value from Intel format to Motorola fomat
 ACCEPTS:     longword value
 RETURNS:     flipped longword value
==========================================================================}
procedure TADFlib.FlipLong(var l: longword);
var
  b1:array[0..3] of byte;
  b2:array[0..3] of byte;

begin
  move(l,b1,4);
  b2[0]:=b1[3];
  b2[1]:=b1[2];
  b2[2]:=b1[1];
  b2[3]:=b1[0];
  move(b2,l,4);
end;


{==========================================================================
 FUNCTION:    calcChecksum
 DESCRIPTION: Calc Checksum of a Block routine written by Feltzkrone
              Dont know how it works yet

 ACCEPTS:     method 0 = bootblock, method 1 = other blocks
 RETURNS:     Checksum that a given block SHOULD have
==========================================================================}
function TADFlib.calcChecksum(var X; XSize:longword; CheckSumPos:longword; Method:longword):longword;
var
  newcarry      :longword;
  carry         :longword;
  newchecksum   :longword;
  l0            :longword;
  l1            :longword;
  CSbuf:array[0..3] of byte;
  Data:array[0..1023] of byte;

begin
  move(x,Data,xSize);
  fillchar(CSBuf, $04, $00);

  if method=1 then
    CSBuf[3]:=0;

  carry:=0;
  newcarry:=0;

  for l0:=$00 to (XSize-1) div 4+1 do { +1 for possible left-over }
  begin
    if (l0 = (XSize-1) div 4+1) then
      for l1:=$03 downto $00 do
      begin
        if CSBuf[l1]+carry>=$100 then
           newcarry:=$01
        else
          newcarry:=$00;

        CSBuf[l1]:= CSBuf[l1] + carry;

        if (method=0) or (l1>0) then
          carry:=newcarry
        else
          carry:=0;
      end
    else
      if l0 <> CheckSumPos div 4 then
        for l1:=$03 downto $00 do
        begin
          if CSBuf[l1]+Data[l0*$04+l1]+carry>=$100 then
            newcarry:=$01
          else
            newcarry:=$00;
               CSBuf[l1] := CSBuf[l1] + Data[l0*$04+l1] + carry;
               if (method=0) or (l1>0) then
            carry:=newcarry
          else
            carry:=0;
        end;
      end;

      move(CSBuf,newchecksum,4);
      if method=0 then
        CalcCheckSum:= not NewCheckSum
      else
      begin
        FlipLong(newchecksum);
        newchecksum:=-newchecksum;
        FlipLong(newchecksum);
        CalcCheckSum:=NewCheckSum;
  end;
end;



end.



